from datetime import datetime
class compare_isolates:

	def compare_isolates(self, isolate1, isolate2):
		reference = ['Complete genome', 'Chromosome', 'Scaffold', 'Contig']
		label = 'assembly_level'
		assembly1 = isolate1[label]
		assembly2 = isolate2[label]
		if assembly1 == assembly2:
			return self.check_paired(isolate1, isolate2)
		else: #change for dictonatry
			for i in reference:
				if i == assembly1:
					return isolate1
				elif i == assembly2:
					return isolate2
			return []


	def check_paired(self, isolate1, isolate2):
		pac_label = 'paired_asm_comp'
		paired1 = isolate1[pac_label]
		paired2 = isolate2[pac_label]
		if paired1 == paired2:
			return self.check_mostrecent(isolate1, isolate2)
		elif paired1 == 'identical':
			return isolate1
		elif paired2 == 'identical':
			return isolate2
		else:
			return -1

	def check_mostrecent(self, isolate1, isolate2):
		label = 'seq_rel_date'
		timeformat = "%Y/%m/%d"
		date1 = datetime.strptime(isolate1[label], timeformat)
		date2 = datetime.strptime(isolate2[label], timeformat)
		if date1 > date2:
			return isolate1
		elif date2 > date1:
			return isolate2
		else:
			return isolate1


# -*- coding: utf-8 -*-
from genome_update import compare_isolates as ci
import os.path
class dictonary:
	def __init__(self, handler, outputfile, inputfile = None):
		self.output_path = outputfile
		self.input_path = inputfile
		self.handler = handler
		self.compare = ci.compare_isolates()

	def insert(self, dictonary, species, isolate, data):	
		dictonary = dictonary
		if species in dictonary:
			selected_species = dictonary[species]
			if isolate not in selected_species:
				print('Adding new isolate %s to species: %s' % (isolate ,species))
				selected_species[isolate] = data
			else:
				selected_isolate = selected_species[isolate]
				new_isolate = self.compare.compare_isolates(selected_isolate, data)
				if len(set(selected_isolate.values()).difference(new_isolate.values())) > 0: 
					print('There is a more complete or newer version of isolate %s in species %s' % (isolate,species))		
		else:
			print('Adding new species: %s to yaml file' %species)
			dictonary[species] = {isolate : data}
		return dictonary


	def create_dict(self, isolate):
		newColumns = ['local_fasta', 'local_gbk', 'local_gff']
		dicto = {}
		indeces = isolate.index
		assembly = 'assembly_accession'
		dicto[assembly] = isolate.name
		for i, j in enumerate(isolate):
			dicto[indeces[i]] = str(j)
		for i in newColumns:
			dicto[i] = ""
		return dicto

	def update_dictonary(self, isolates,  genus):
		local = {}
		if (self.input_path != None):
			local = self.handler.read(self.input_path)
		
		for i in isolates:
			dicto = self.create_dict(i[2])
			local = self.insert(local, genus +' '+ i[0], i[1], dicto)
		self.handler.write(self.output_path,local)




# -*- coding: utf-8 -*-
import pandas as pd
from genome_update import compare_isolates as ci

class isolate_finder:

	def __init__(self, assembly_path):
		
		self.frame = self.load_assemblyinfo(assembly_path)
		self.labels = self.find_labels()
		self.compare = ci.compare_isolates()


	def load_assemblyinfo(self,url):
		genbankout = pd.read_table(url, sep='\t', header=1, index_col=0)
		return genbankout

	def find_labels(self):
		indices = {}
		for index, name in enumerate(self.frame.columns.values):
			indices[name] = index
		return indices


	def select_genus(self, target, species_id = None, tax_id = None):
		indices = []
		unique = []
		for i in range(len(self.frame.index)):
			genus = self.frame.iat[i, self.labels['organism_name']].split(' ')
			if target in genus:
				isolate = self.select_isolate(genus[1], i)
				if species_id is not None and species_id != str(self.frame.iat[i, self.labels['species_taxid']]):
					continue
				if tax_id is not None and tax_id != str(self.frame.iat[i, self.labels['taxid']]):
					print('hey from tax')
					continue 
				if isolate not in unique:
					indices.append(i)
					unique.append(isolate)
				else:
					oldIsolate = unique.index(isolate)
					oldIsolateIndex = self.frame.iloc[indices[oldIsolate]] 
					selectedIndex = self.compare.compare_isolates(oldIsolateIndex, self.frame.iloc[i])
					if not selectedIndex.empty:
						newindex = self.frame.index.get_loc(selectedIndex.name)
						indices[oldIsolate] = newindex
						unique[oldIsolate] = self.select_isolate(genus[1], newindex)
		return self.return_isolates(unique, indices)

	def return_isolates(self, names ,indices):
		returnList = []
		for i, j in enumerate(indices):
			tempList = names[i]
			tempList.append(self.frame.iloc[j])
			returnList.append(tempList)
		return returnList

	def select_isolate(self, species, index):
		isolate_name = self.get_isolate_name(index)
		isolate = [species, isolate_name]
		return isolate

	def get_isolate_name(self, index):
		name = self.frame.iat[index, self.labels['infraspecific_name']]
		if pd.isnull(name):
			name = self.frame.iat[index, self.labels['isolate']]
			if pd.isnull(name):
				genus = self.frame.iat[index, self.labels['organism_name']].split(' ')
				if len(genus) > 2:
					name = ' '.join(genus[2:])
				else:
					name = self.frame.iat[index, self.labels['asm_name']]
		else:
			name = name.replace('strain=','')
		return name




import yaml

class yaml_handler:

	def write(self, file, data):
		with open(file,'w') as yam:
			yaml.dump(data,yam, default_flow_style=False)

	def read(self, file):
		with open(file,'r') as yam:
			return next(yaml.load_all(yam))


# -*- coding: utf-8 -*-
import urllib
import os
from multiprocessing import Pool

class downloader:
	def __init__(self, genome):
		self.genomes = genome
		self.file_labels = ['local_fasta', 'local_gbk', 'local_gff']
		self.file_formats = ['fna.gz', 'gbff.gz', 'gff.gz']
	def get_download_jobs(self, outfolder):
		jobs = []
		for species in self.genomes.keys():
			for strain in self.genomes[species].keys():
				for j, i in enumerate(self.file_labels):
					path = self.genomes[species][strain][i]
					if path == '' or path == None:
						assembly = self.genomes[species][strain]['assembly_accession']
						asm_name = self.genomes[species][strain]['asm_name'].replace(' ','_').replace('/',"_")
						ftp_path = self.genomes[species][strain]['ftp_path']
						ftp_path = '/'.join([ftp_path , ftp_path[ftp_path.find(assembly):]])
						ftp_path = ftp_path +'_'+ 'genomic' +'.'+ self.file_formats[j]
						file_name = os.path.join(outfolder, strain.replace('/','_') + "." + self.file_formats[j])
						jobs.append([ftp_path, file_name, species, strain, i])
		return jobs

	def download(self, job):
		try:
			urllib.request.urlretrieve(job[0], job[1])
			print('Download of %s complete' %job[1])
		except urllib.error.HTTPError:
			print('Download of %s failed could not find ftp: '%job[0])

	def download_jobs(self, parallel, jobs):
		print('Downloading jobs....')
		pool = Pool(processes=parallel)
		pool.map(self.download, jobs)
		print('Download complete')
		return self.check_files(jobs)

	def check_files(self, jobs):
		for job in jobs:
			if os.path.isfile(job[1]):
				self.genomes[job[2]][job[3]][job[4]] = job[1]
		return self.genomes


# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
	name='genome_update',
	version = '0.1.5',
	description = 'A script for updating a local directory with genomes, from NCBI assembly files',
	long_description = long_description,
	url='https://github.com/Emisam/genome_update.git',
	author='Emil Samuelsson',
	author_email='Emil@Samuelsson.pp.se',
	licence='MIT',
	classifiers=[
		'Development Status :: 3 - Alpha',
		'Intended Audience :: Science/Research',
		'Topic :: Scientific/Engineering :: Bio-Informatics',
		'License :: OSI Approved :: MIT License',
		'Programming Language :: Python :: 3',
	],
	packages=find_packages(),
	install_requires=['pandas', 'PyYAML'],
	entry_points={
        'console_scripts': [
            'genome_update=genome_update.__main__:main'
        ]
    }
	)


# -*- coding: utf-8 -*-
import argparse
import os

import genome_update
from genome_update import isolate_finder as iso
from genome_update import yaml_handler as yh
from genome_update import dictonary_handler as dh
from genome_update import downloader as dw
def main():
	SUPPORTED_DOMAINS = ['archaea', 'bacteria', 'fungi', 'invertebrate', 'plant', 'protozoa', 'vertebrate_mammalian', 'vertebrate_other', 'viral']
	parser = argparse.ArgumentParser()

	parser.add_argument('-name', '--newname', 
						dest = 'name',
						default = 'genome.yaml',
						help = 'What the yaml file should be named')
	parser.add_argument('-o', '--output',
						dest = 'output', 
						default = 'genomes/', 
						help = 'The output which files will be downloaded to')
	parser.add_argument('-p', '--parallel', 
						dest = 'parallel', 
						default = 1, 
						type = int, 
						metavar = "N", 
						help = 'Use N processes to download files (default = 1)')
	parser.add_argument('-i', '--input-file-path', 
						dest = 'input', 
						default = None, 
						help = 'Local yaml file to update')
	
	parser.add_argument('-domain','--domain', 
						dest = 'domain', 
						choices = SUPPORTED_DOMAINS, 
						default = 'bacteria',
						help = 'The domain to download from')
	parser.add_argument('-g', '--genus', 
						dest = 'genus', 
						help = 'Specify the target genus' )
	parser.add_argument('-s', '--species-taxid', 
						dest = 'species_taxid',
						default = None, 
						help = 'Used to specify a target species_taxid' )
	parser.add_argument('-t', '--taxid', 
						dest = 'taxid', 
						default = None, 
						help = 'Used to specify a target taxid')

	parser.add_argument('-d', '--download', 
						dest = 'download', 
						help = 'Use to only download files for all unique isolates in a yaml file', 
						action = 'store_true')
	parser.add_argument('-u', '--update-yaml', 
						dest = 'update', 
						help = 'Use to only update local yaml file', 
						action = 'store_true')
	parser.add_argument('-V', '--version', action ='version', version=genome_update.__version__, help='print version information')

	args = parser.parse_args()


	NCBI = 'ftp://ftp.ncbi.nlm.nih.gov/genomes/genbank/' + args.domain + '/assembly_summary.txt'
	yaml = yaml_handler()

	if not os.path.exists(args.output):
		os.makedirs(args.output)

	if args.download and args.update:
		print('Both --download and --update-yaml cannot be supplied')
	elif args.download:
		if args.input != None and os.path.exists(args.input):
			download(yaml, args.input, args.name, args.output, args.parallel)
		else:
			print('To use --download, specify a correct input with -i')
	elif args.update:
		if args.input != None and args.genus and os.path.exists(args.input):
			print('Checking if input is uptodate with NCBI assembly')
			update_yaml(yaml, NCBI, args.genus, args.species_taxid, args.taxid, args.name, args.input)
		else:
			print('To use --download, specify a correct input with -i')
	elif args.genus:
		update_yaml(yaml, NCBI, args.genus, args.species_taxid, args.taxid, args.name, args.input)
		download(yaml, args.name, args.name, args.output, args.parallel)
	else:
		print('Input arguments was invalid, try -h for help')




def download(yaml, inputfile, outputfile, directory, parallel):
	dwon = dw.downloader(yaml.read(inputfile))
	job = dwon.get_download_jobs(directory)
	if len(job) != 0:
		newdict = dwon.download_jobs(parallel, job)
		yaml.write(outputfile, newdict)
	else:
		print('Could not find any jobs, check if genus and species-taxid is correctly entered')

def update_yaml(yaml, url, genus, species_taxid, taxid, outputfile, inputfile):
	dl = iso.isolate_finder(url)
	isolates = dl.select_genus(genus, species_taxid, taxid)
	dictonary = dh.dictonary(yaml, outputfile, inputfile)
	dictonary.update_dictonary(isolates, genus)

if __name__ == '__main__':
	main()